@@ -2,22 +2,26 @@
 Functions/classes/variables for interacting between a pandas DataFrame
 and postgres/mysql/sqlite (and potentially other databases).
 """
-import json
 import pandas as pd
 import logging
 import re
+import sqlalchemy as sa
 from copy import deepcopy
 from distutils.version import LooseVersion
 from math import floor
 from sqlalchemy import JSON, MetaData, select
 from sqlalchemy.sql import null
-from sqlalchemy.schema import (PrimaryKeyConstraint, CreateColumn, CreateSchema)
+from sqlalchemy.schema import PrimaryKeyConstraint, CreateSchema
 from alembic.runtime.migration import MigrationContext
 from alembic.operations import Operations
 from pangres.logger import log
-from pangres.upsert import (mysql_upsert,
-                            postgres_upsert,
-                            sqlite_upsert)
+from pangres.exceptions import (BadColumnNamesException,
+                                DuplicateLabelsException,
+                                DuplicateValuesInIndexException,
+                                HasNoSchemaSystemException,
+                                MissingIndexLevelInSqlException,
+                                UnnamedIndexLevelsException)
+from pangres.upsert import UpsertQuery
 
 # # Regexes
 
@@ -31,6 +35,7 @@ RE_CHARCOUNT_COL_TYPE = re.compile(r'(?<=.)+\(\d+\)')
 
 # # "Adapter" for sqlalchemy (handle pre and post version 1.4)
 
+# +
 def _sqla_gt14() -> bool:
     """
     Checks if sqlalchemy.__version__ is at least 1.4.0, when several
@@ -43,6 +48,18 @@ def _sqla_gt14() -> bool:
     return LooseVersion(sqlalchemy.__version__) >= LooseVersion("1.4.0")
 
 
+def _sqlite_gt3_22_0() -> bool:
+    """
+    Checks if the SQLite version is >= than 3.22.0.
+    Starting from this version we can use more SQL parameters.
+    See https://github.com/ThibTrip/pangres/issues/43
+    """
+    import sqlite3
+    return LooseVersion(sqlite3.sqlite_version) >= LooseVersion("3.22.0")
+
+
+# -
+
 # # Class PandasSpecialEngine
 
 class PandasSpecialEngine:
@@ -85,7 +102,7 @@ class PandasSpecialEngine:
             Similar to pd.to_sql dtype argument.
             This is especially useful for MySQL where the length of
             primary keys with text has to be provided (see Examples)
-        
+
         Examples
         --------
         >>> from sqlalchemy import create_engine
@@ -114,17 +131,19 @@ class PandasSpecialEngine:
         if self._db_type == "postgres":
             schema = 'public' if schema is None else schema
             # raise if we find columns with "(", ")" or "%"
-            if any((RE_BAD_COL_NAME.search(col) for col in df.columns)):
-                err = ("psycopg2 (Python postgres driver) does not seem to support"
+            bad_col_names = [col for col in df.columns if RE_BAD_COL_NAME.search(col)]
+            if len(bad_col_names) > 0:
+                err = ("psycopg2 (Python postgres driver) does not seem to support "
                        "column names with '%', '(' or ')' "
-                       "(see https://github.com/psycopg/psycopg2/issues/167)")
-                raise ValueError(err)
+                       "(see https://github.com/psycopg/psycopg2/issues/167). You need to fix "
+                       f"these names: {bad_col_names}")
+                raise BadColumnNamesException(err)
 
         # VERIFY ARGUMENTS
         # all index levels have names
         index_names = list(df.index.names)
-        if any([ix_name is None for ix_name in index_names]):
-            raise IndexError("All index levels must be named!")
+        if any(ix_name is None for ix_name in index_names):
+            raise UnnamedIndexLevelsException("All index levels must be named!")
 
         # index is unique
         if not df.index.is_unique:
@@ -133,13 +152,14 @@ class PandasSpecialEngine:
                    "Check duplicates using this code (assuming df "
                    " is the DataFrame you want to upsert):\n"
                    ">>> df.index[df.index.duplicated(keep=False)]")
-            raise IndexError(err)
+            raise DuplicateValuesInIndexException(err)
 
         # there are no duplicated names
         fields = list(df.index.names) + df.columns.tolist()
         if len(set(fields)) != len(fields):
-            raise ValueError(("There cannot be duplicated names amongst "
-                              "index levels and/or columns!"))
+            duplicated_labels = [c for c in fields if fields.count(c) > 1]
+            raise DuplicateLabelsException("Found duplicates across index "
+                                           f"and columns: {duplicated_labels}")
 
         # detect json columns
         def is_json(col):
@@ -168,7 +188,6 @@ class PandasSpecialEngine:
         constraint = PrimaryKeyConstraint(*[table.columns[name]
                                             for name in df.index.names])
         table.append_constraint(constraint)
-        
 
         # add remaining attributes
         self.engine = engine
@@ -176,7 +195,6 @@ class PandasSpecialEngine:
         self.schema = schema
         self.table = table
 
-
     @staticmethod
     def _detect_db_type(engine) -> str:
         """
@@ -197,6 +215,24 @@ class PandasSpecialEngine:
         else:
             return "other"
 
+    def _raise_no_schema_feature(self):
+        """
+        Function to raise an error if we try to do operations on schemas on a database that does not
+        support such feature.
+        """
+        # Should I just do self.db_type != 'postgres'? (not sure if any other DBs use schemas)
+        if self._db_type not in ('postgres', 'other'):
+            raise HasNoSchemaSystemException('Cannot create schemas for given SQL flavor '
+                                             '(AFAIK only PostgreSQL has this feature)')
+
+    def schema_exists(self, connection) -> bool:
+        self._raise_no_schema_feature()
+        if _sqla_gt14():
+            insp = sa.inspect(connection)
+            return self.schema in insp.get_schema_names()
+        else:
+            return self.engine.dialect.has_schema(connection, self.schema)
+
     def table_exists(self) -> bool:
         """
         Returns True if the table defined in given instance
@@ -219,9 +255,20 @@ class PandasSpecialEngine:
         Creates the schema defined in given instance of
         PandasSpecialEngine if it does not exist.
         """
-        if (self.schema is not None and
-            not self.engine.dialect.has_schema(self.engine, self.schema)):
-            self.engine.execute(CreateSchema(self.schema))
+        self._raise_no_schema_feature()
+        if self.schema is None:
+            raise AssertionError('Cannot create schema because it is None. '
+                                 'If you used PostgreSQL the schema should have defaulted to `public`. '
+                                 'If using something other than PostgreSQL make sure it supports schemas '
+                                 '(AFAIK only PostgreSQL has this feature) and retry your operation '
+                                 'after setting the default schema `public` yourself '
+                                 '(if that is the schema you wish to use).')  # pragma: no cover
+
+        with self.engine.connect() as connection:
+            if not self.schema_exists(connection=connection):
+                connection.execute(CreateSchema(self.schema))
+                if hasattr(connection, 'commit'):
+                    connection.commit()
 
     def create_table_if_not_exists(self):
         """
@@ -241,7 +288,6 @@ class PandasSpecialEngine:
             list of column names (str)
         """
         if _sqla_gt14():
-            import sqlalchemy as sa
             insp = sa.inspect(self.engine)
             columns_info = insp.get_columns(schema=self.schema, table_name=self.table.name)
         else:
@@ -267,9 +313,9 @@ class PandasSpecialEngine:
                        if col.name not in self.get_db_columns_names()]
         # check columns are not index levels
         if any((c.name in self.df.index.names for c in cols_to_add)):
-            raise ValueError(('Cannot add any column that is part of the df index!\n'
-                              "You'll have to update your table primary key or change your "
-                              "df index"))
+            raise MissingIndexLevelInSqlException('Cannot add any column that is part of the df index!\n'
+                                                  "You'll have to update your table primary key or change your "
+                                                  "df index")
 
         with self.engine.connect() as con:
             ctx = MigrationContext.configure(con)
@@ -279,7 +325,8 @@ class PandasSpecialEngine:
                 op.add_column(self.table.name, col, schema=self.schema)
                 log(f"Added column {col} (type: {col.type}) in table {self.table.name} "
                     f'(schema="{self.schema}")')
-
+            if hasattr(con, 'commit'):
+                con.commit()
 
     def get_db_table_schema(self):
         """
@@ -294,7 +341,7 @@ class PandasSpecialEngine:
         table_name = self.table.name
         schema = self.schema
         engine = self.engine
-        
+
         metadata = MetaData(bind=engine, schema=schema)
         metadata.reflect(bind=engine, schema=schema, only=[table_name])
         namespace = table_name if schema is None else f'{schema}.{table_name}'
@@ -320,7 +367,8 @@ class PandasSpecialEngine:
             stmt = select(from_obj=db_table,
                           columns=[col],
                           whereclause=col.isnot(None)).limit(1)
-            results = self.engine.execute(stmt).fetchall()
+            with self.engine.connect() as connection:
+                results = connection.execute(stmt).fetchall()
             if results == []:
                 empty_columns.append(col)
         return empty_columns
@@ -384,6 +432,8 @@ class PandasSpecialEngine:
                     log(f"Changed type of column {new_col.name} "
                         f"from {col.type} to {new_col.type} "
                         f'in table {self.table.name} (schema="{self.schema}")')
+                    if hasattr(con, 'commit'):
+                        con.commit()
 
     @staticmethod
     def _create_chunks(values, chunksize=10000):
@@ -431,11 +481,53 @@ class PandasSpecialEngine:
                     values[i][j] = null()
                 # cast pd.Interval to str
                 elif isinstance(val, pd.Interval):
-                    warnings.warn(('found pd.Interval objects, '
-                                   'they will be casted to str'))
+                    log('found pd.Interval objects, they will be casted to str',
+                        level=logging.WARNING)
                     values[i][j] = str(val)
         return values
-    
+
+    def _sqlite_chunsize_fix(self, chunksize):
+        """
+        Changes the chunksize given by the user to circumvent the max of 999 parameters
+        for sqlite.
+
+        Raises
+        ------
+        NotImpletementedError
+            When tables have more than 999 columns.
+            In such a case even single row inserts have already too many variables.
+            (you can google "SQLITE_MAX_VARIABLE_NUMBER" for more info)
+
+        Examples
+        --------
+        >>> from sqlalchemy import create_engine
+        >>>
+        >>> # this assumes you have SQlite version >= 3.22.0 
+        >>> engine = create_engine("sqlite:///:memory:)")
+        >>> df = pd.DataFrame({'name':['Albert']}).rename_axis(index='profileid')
+        >>> pse = PandasSpecialEngine(engine=engine, df=df, table_name='example')
+        >>> pse._sqlite_chunsize_fix(chunksize=33000)
+        16383
+        """
+        maximum = 32766 if _sqlite_gt3_22_0() else 999
+        new_chunksize = floor(maximum / len(self.table.columns))
+        # note: the part below cannot be realistically tested as the table creation
+        # with `pd.io.sql.SQLTable` just takes too much time in presence of a table
+        # with many many columns
+        if new_chunksize < 1:  # pragma: no cover
+            # case > maximum columns
+            err = (f'Updating SQlite tables with more than {maximum} columns is '
+                   f'not supported due to max variables restriction ({maximum} max).\n'
+                   'If you know a way around that please let me know '
+                   '(e.g. post a GitHub issue)!')
+            raise NotImplementedError(err)
+        if chunksize > new_chunksize:
+            log(f'Reduced chunksize from {chunksize} to {new_chunksize} due '
+                f'to SQlite max variable restriction (max {maximum}).',
+                level=logging.WARNING)
+            chunksize = new_chunksize
+        return chunksize
+
     def upsert(self, if_row_exists, chunksize=10000, yield_chunks=False):
         """
         Generates and executes an upsert (insert update or 
@@ -463,55 +555,27 @@ class PandasSpecialEngine:
             (sqlalchemy.engine.cursor.LegacyCursorResult)
             at each chunk with which you can for instance count rows.
         """
-        # VERIFY ARGUMENTS
-        if if_row_exists not in ('ignore', 'update'):
-            raise ValueError('if_row_exists must be "ignore" or "update"')
+        assert if_row_exists in ('ignore', 'update')
         # convert values if needed
         values = self._get_values_to_insert()
         # recalculate chunksize for sqlite
         if self._db_type == 'sqlite':
-            # to circumvent the max of 999 parameters for sqlite we have to
-            # make sure chunksize is not too high also I don't know how to
-            # deal with tables that have more than 999 columns because even
-            # with single row inserts it's already too many variables.
-            # (you can google "SQLITE_MAX_VARIABLE_NUMBER" for more info)
-            new_chunksize = floor(999 / len(self.table.columns))
-            if new_chunksize < 1:
-                # case > 999 columns
-                err = ('Updating SQlite tables with more than 999 columns is '
-                       'not supported due to max variables restriction (999 max). '
-                       'If you know a way around that please let me know'
-                       '(e.g. post a GitHub issue)!')
-                raise NotImpletementedError(err)
-            if chunksize > new_chunksize:
-                log(f'Reduced chunksize from {chunksize} to {new_chunksize} due '
-                    'to SQlite max variable restriction (max 999).',
-                    level=logging.WARNING)
-                chunksize = new_chunksize
+            chunksize = self._sqlite_chunsize_fix(chunksize=chunksize)
         # create chunks
         chunks = self._create_chunks(values=values, chunksize=chunksize)
-        upsert_funcs = {"postgres":postgres_upsert,
-                        "mysql":mysql_upsert,
-                        "sqlite":sqlite_upsert,
-                        "other":sqlite_upsert}
-        upsert_func = upsert_funcs[self._db_type]
+        upq = UpsertQuery(engine=self.engine, table=self.table)
+
         # when yield is present in a function it always returns a generator
         # even if the yield is at a place where the code is not supposed to execute
         # but one can circumvent this by using a subfunction
         # which is what we do for when we want to yield results of chunks
         if not yield_chunks:
             for chunk in chunks:
-                upsert_func(engine=self.engine,
-                            table=self.table,
-                            values=chunk,
-                            if_row_exists=if_row_exists)
+                upq.execute(db_type=self._db_type, values=chunk, if_row_exists=if_row_exists)
         else:
             def yield_chunks_func():
                 for chunk in chunks:
-                    yield upsert_func(engine=self.engine,
-                                      table=self.table,
-                                      values=chunk,
-                                      if_row_exists=if_row_exists)
+                    yield upq.execute(db_type=self._db_type, values=chunk, if_row_exists=if_row_exists)
             return yield_chunks_func()
 
     def __repr__(self):
@@ -521,7 +585,7 @@ class PandasSpecialEngine:
                    * table: {self.table.name}
                    * SQLalchemy table model:\n{self.table.__repr__()}"""
         text = '\n'.join([line.strip() for line in text.splitlines()])
-        
+
         df_repr = (str(self.df.head()) if not hasattr(self.df, 'to_markdown')
                    else str(self.df.head().to_markdown()))
         text += f'\n* df.head():\n{df_repr}'
