@@ -1,5 +1,6 @@
 import pandas as pd
 from pangres.logger import log
+from pangres.exceptions import DuplicateLabelsException, UnnamedIndexLevelsException
 
 
 # # Function to fix bad column names for psycopg2
@@ -24,6 +25,18 @@ def fix_psycopg2_bad_cols(df:pd.DataFrame, replacements={'%':'', '(':'', ')':''}
         The keys '%', '(' and ')' are mandatory.
         There cannot be any extra keys.
 
+    Raises
+    ------
+    pangres.exceptions.UnnamedIndexLevelsException
+        When you pass a df where not all index levels are named
+    pangres.exceptions.DuplicateLabelsException
+        When you pass a df with duplicated labels accross index/columns
+        or when after cleaning we end up with duplicated labels
+        e.g. "test(" and "test)" would by default both be renamed to "test"
+    TypeError
+        When `replacements` is not of the expected type or has wrong keys
+        or has non string values
+
     Returns
     -------
     new_df : pd.DataFrame
@@ -31,6 +44,7 @@ def fix_psycopg2_bad_cols(df:pd.DataFrame, replacements={'%':'', '(':'', ')':''}
     Examples
     --------
     * fix bad col/index names with default replacements (empty string for '(', ')' and '%')
+    >>> from pangres import fix_psycopg2_bad_cols
     >>> import pandas as pd
     >>> df = pd.DataFrame({'test()':[0],
     ...                    'foo()%':[0]}).set_index('test()')
@@ -63,35 +77,40 @@ def fix_psycopg2_bad_cols(df:pd.DataFrame, replacements={'%':'', '(':'', ')':''}
     # verify all index levels are named
     index_names = list(df.index.names)
     if any([ix_name is None for ix_name in index_names]):
-        raise IndexError("All index levels must be named!")
+        raise UnnamedIndexLevelsException("All index levels must be named!")
+
     # verify duplicated columns
     fields = list(df.index.names) + df.columns.tolist()
     if len(set(fields)) != len(fields):
-        raise ValueError(("There cannot be duplicated names amongst "
-                          "index levels and/or columns!"))
+        duplicates = [c for c in fields if fields.count(c) > 1]
+        raise DuplicateLabelsException("There cannot be duplicated names amongst "
+                                       f"index levels and/or columns! Duplicates found: {duplicates}")
     # verify replacements arg
     expected_keys = ('%', '(', ')')
     if ((not isinstance(replacements, dict)) or
-        (set(replacements.keys()) - set(expected_keys) != set()) or 
+        (set(replacements.keys()) - set(expected_keys) != set()) or
         (len(replacements) != len(expected_keys))):
         raise TypeError(f'replacements must be a dict containing the following keys (and none other): {expected_keys}')
     if not all((isinstance(v, str) for v in replacements.values())):
-        raise TypeError(f'The values of replacements must all be strings')
+        raise TypeError('The values of replacements must all be strings')
+
     # replace bad col names
     translator = {ord(k):v for k, v in replacements.items()}
     new_df = df.copy(deep=True)
     renamer = lambda col: col.translate(translator) if isinstance(col, str) else col
     new_df = new_df.rename(columns=renamer).rename_axis(index=renamer)
+
     # check columns are unique after renaming
     fields = list(new_df.index.names) + new_df.columns.tolist()
     if len(set(fields)) != len(fields):
-        raise ValueError("Columns/index are not unique after renaming!")
+        duplicates = [c for c in fields if fields.count(c) > 1]
+        raise DuplicateLabelsException("Columns/index are not unique after renaming! "
+                                       f"Duplicates found: {duplicates}")
+
     # compare columns (and index)
     before = df.reset_index().columns.tolist()
     after = new_df.reset_index().columns.tolist()
-    changed = []
     for i, j in zip(before, after):
-        if (isinstance(i, str) and isinstance(j, str)
-            and i!=j):
+        if (isinstance(i, str) and isinstance(j, str) and i != j):
             log(f'Renamed column/index "{i}" to "{j}s"')
     return new_df
